import io
import os
import pprint
import re
from json.decoder import JSONDecodeError

import boto3
import logging
import requests
from boto3.s3.transfer import TransferConfig
from rho_ml.stored_model import StoredModel
from typing import Dict, Type, Optional

from rho_ml.rho_model import RhoModel
from rho_ml.serialization import LocalModelStorage
from rho_ml.utils import Version

from sermos_utils.constants import DEFAULT_STORE_MODEL_URL, S3_MODEL_BUCKET, \
    DEFAULT_GET_MODEL_URL

logger = logging.getLogger(__name__)


def get_storage_key(model: Type[RhoModel], prefix: str) -> str:
    return "{0}/{1}_{2}".format(prefix, model.name, model.version)


def get_headers(api_key: str) -> Dict[str, str]:
    return {
        'Content-Type': 'application/json',
        'apikey': api_key
    }


def get_s3_client_from_api_response(response_data: Dict[str, str]):
    try:
        access_key = response_data['aws_access_key']
        secret_key = response_data['aws_secret_key']
        session_token = response_data['aws_session_token']
        region = response_data['aws_region']
    except KeyError as e:
        logger.warning("Missing IAM keys in response:\n{0}"
                       .format(pprint.pformat(response_data)))
        raise e
    else:
        client = boto3.client('s3',
                              aws_access_key_id=access_key,
                              aws_secret_access_key=secret_key,
                              aws_session_token=session_token,
                              region_name=region)
        return client


# todo: come up w/ way to compare hashes of stored models and loaded models
def store_rho_model(model: Type[RhoModel],
                    deploy_key: Optional[str] = None,
                    store_model_endpoint: Optional[str] = None,
                    base_class_name: Optional[str] = None,
                    transfer_config: Optional[TransferConfig] = None):
    """ Get S3 credentials from the sermos-admin API, then use the credentials
     to store the model and metadata in the appropriate bucket / subfolder"""
    logger.info("Attempting to serialize {0}".format(model.name))
    stored_model = StoredModel.from_model(model=model,
                                          base_class_name=base_class_name)
    store_bytes = stored_model.to_pickle()
    f = io.BytesIO()
    f.write(store_bytes)
    f.seek(0)
    logger.info("Serialization finished, requesting credentials from API...")
    if not deploy_key:
        deploy_key = os.environ['SERMOS_DEPLOY_KEY']
    headers = get_headers(api_key=deploy_key)
    if not store_model_endpoint:
        store_model_endpoint = DEFAULT_STORE_MODEL_URL
    post_data = {'model_key': model.name + '_' + str(model.version)}
    r = requests.post(url=store_model_endpoint, headers=headers, json=post_data)
    response_data = r.json()
    client = get_s3_client_from_api_response(response_data=response_data)
    storage_key = response_data['model_key']
    logger.info("Uploading serialized {0}"
                .format('_'.join([model.name, str(model.version)])))
    client.upload_fileobj(
        f, Bucket=S3_MODEL_BUCKET, Key=storage_key, Config=transfer_config)


class ModelNotFoundError(Exception):
    pass


def get_model_search_api_response(model_name: str,
                                  version_pattern: str,
                                  deploy_key: str,
                                  get_model_endpoint: Optional[str] = None) \
        -> Dict[str, str]:
    headers = get_headers(api_key=deploy_key)
    if not get_model_endpoint:
        get_model_endpoint = DEFAULT_GET_MODEL_URL
    query_params = {
        'model_name': model_name,
        'version_pattern': version_pattern
    }
    logger.debug("Requesting storage info from Admin API...")
    r = requests.post(
        url=get_model_endpoint, headers=headers, json=query_params)
    try:
        response_data = r.json()
    except JSONDecodeError:
        logger.warning(f"Could not decode search API response!  "
                       f"Server response: {r.content}")
        raise ModelNotFoundError(f"Invalid response received from admin API!\n"
                                 f" {r.content}")
    except Exception as e:
        logger.exception(e)
        raise e

    return response_data


def get_stored_model_json_from_api_response(response_data: Dict[str, str]) \
        -> str:
    client = get_s3_client_from_api_response(response_data=response_data)
    model_key = response_data['model_key']
    if not model_key:
        raise ModelNotFoundError("No model key found!")
    logger.debug("Retrieving {0} from storage...".format(model_key))
    s3_response = client.get_object(Bucket=S3_MODEL_BUCKET, Key=model_key)
    model_json = s3_response['Body'].read()
    return model_json


def get_local_storage(base_path: Optional[str]) -> LocalModelStorage:
    if base_path:
        result = LocalModelStorage(base_path=base_path)
    else:
        result = LocalModelStorage()
    return result


def get_local_model(model_name: str,
                    version_pattern: str,
                    local_base_path: str) -> Optional[RhoModel]:
    """ Search for a local model by name and version and instantiate it
    if found. """
    local_storage = get_local_storage(base_path=local_base_path)
    local_key = local_storage.get_key_from_pattern(
        model_name=model_name, version_pattern=version_pattern)
    if local_key:
        logger.info("Local model found with filename: {0}".format(local_key))
        result = local_storage.retrieve(local_key)
        return result


def get_version_from_key(key: str) -> Version:
    m = re.search(r'[0-9]+\.[0-9]+\.[0-9]+', key)
    if m:
        version_string = m.group()
        result = Version.from_string(version_string)
        return result
    else:
        raise ValueError(f"Couldn't create Version from {key}!")


def rho_model_loader(model_name: str,
                     version_pattern: str,
                     deploy_key: Optional[str] = None,
                     local_base_path: Optional[str] = None,
                     save_to_local_disk: bool = True,
                     get_model_endpoint: Optional[str] = None,
                     force_search: bool = False,
                     base_class_name: Optional[str] = None) -> Type[RhoModel]:
    """ Retrieve models stored using the Sermos Admin API by name and
    version.

    Optionally cache the result locally for later use (caching on by
    default). If force_search == True, always check cloud storage,
    regardless of whether or not """
    if not deploy_key:
        deploy_key = os.environ['SERMOS_DEPLOY_KEY']
    if force_search:
        api_response = get_model_search_api_response(
            model_name=model_name, version_pattern=version_pattern,
            deploy_key=deploy_key, get_model_endpoint=get_model_endpoint)
        model_key = api_response['model_key']
        remote_version = get_version_from_key(key=model_key)

    force_get_remote = False
    model = get_local_model(model_name=model_name,
                            version_pattern=version_pattern,
                            local_base_path=local_base_path)
    if model and force_search:
        if model.version < remote_version:
            force_get_remote = True
    if (not model) or force_get_remote:
        if not force_search:
            api_response = get_model_search_api_response(
                model_name=model_name, version_pattern=version_pattern,
                deploy_key=deploy_key, get_model_endpoint=get_model_endpoint)
        stored_bytes = get_stored_model_json_from_api_response(
            response_data=api_response)
        stored_model = StoredModel.from_pickle(stored_bytes)
        model = stored_model.load_model()
        if save_to_local_disk:
            logger.info("Caching retrieved model {0} locally..."
                        .format(model.name))
            local_storage = get_local_storage(base_path=local_base_path)
            local_storage.store(model=model, base_class_name=base_class_name)
    return model



""" Constants used throughout utils.
"""
from urllib.parse import urljoin

ENV_VAR_DEPLOY_KEY = "SERMOS_DEPLOY_KEY"
ENV_VAR_PKG_NAME = "SERMOS_CLIENT_PKG_NAME"
DEFAULT_BASE_URL = "https://admin.sermos.ai/api/v1/"
DEFAULT_DEPLOY_URL = urljoin(DEFAULT_BASE_URL, 'deploy')
DEFAULT_GET_MODEL_URL = urljoin(DEFAULT_BASE_URL, 'models/get-model/')
DEFAULT_STORE_MODEL_URL = urljoin(DEFAULT_BASE_URL, 'models/store-model/')
DEFAULT_YAML_NAME = "sermos.yaml"
S3_MODEL_BUCKET = 'sermos-client-models'

# Used to define the 'name' of the default, base image that is used for
# every customer build unless they define imageConfig.customImages in their
# sermos.yaml and then specify the customImageName key on either the API
# or custom workers.
DEFAULT_BASE_IMAGE_NAME = 'base'



""" Generic utilities
"""
import os
import logging
from sermos_utils.constants import ENV_VAR_DEPLOY_KEY, ENV_VAR_PKG_NAME

logger = logging.getLogger(__name__)


def normalized_pkg_name(pkg_name: str, dashed: bool = False):
    """ We maintain consistency by always specifying the package name as
        the "dashed version".

        Python/setuptools will replace "_" with "-" but resource_filename()
        expects the exact directory name, essentially. In order to keep it
        simple upstream and *always* provide package name as the dashed
        version, we do replacement here to 'normalize' both versions to
        whichever convention you need at the time.

        if `dashed`:
            my-package-name --> my-package-name
            my_package_name --> my-package-name

        else:
            my-package-name --> my_package_name
            my_package_name --> my_package_name
    """
    if dashed:
        return str(pkg_name).replace('_', '-')
    return str(pkg_name).replace('-', '_')


def get_deploy_key(deploy_key: str = None):
    """ Verify deploy key provided, get from environment if None.

        Raise if neither provided nor found.

        Arguments:
            deploy_key (optional): Deployment key, issued by Sermos, that
                dictates the environment into which this request will be
                deployed. Defaults to checking the environment for
                `SERMOS_DEPLOY_KEY`. If not found, will exit.
    """
    deploy_key = deploy_key if deploy_key\
        else os.environ.get(ENV_VAR_DEPLOY_KEY, None)
    if deploy_key is None:
        msg = "Unable to find `deploy-key` in CLI arguments nor in "\
            "environment under `{}`".format(ENV_VAR_DEPLOY_KEY)
        logger.error(msg)
        raise ValueError(msg)
    return deploy_key


def get_client_pkg_name(pkg_name: str = None):
    """ Verify the package name provided and get from environment if None.

        Raise if neither provided nor found.

        Arguments:
          pkg_name (optional): Directory name for your Python
                    package. e.g. my_package_name . If none provided, will check
                    environment for `SERMOS_CLIENT_PKG_NAME`. If not found,
                    will exit.
    """
    pkg_name = pkg_name if pkg_name\
        else os.environ.get(ENV_VAR_PKG_NAME, None)
    if pkg_name is None:
        msg = "Unable to find `pkg-name` in CLI arguments nor in "\
            "environment under `{}`".format(ENV_VAR_PKG_NAME)
        logger.error(msg)
        raise ValueError(msg)
    return pkg_name



""" Definition of the `sermos.yaml` file
"""
import re
import os
import logging
import pkg_resources
import yaml
from yaml.loader import SafeLoader
from marshmallow import Schema, fields, pre_load, EXCLUDE, INCLUDE
from marshmallow.exceptions import ValidationError
from sermos_utils.utils import normalized_pkg_name
from sermos_utils.constants import DEFAULT_YAML_NAME, DEFAULT_BASE_IMAGE_NAME

logger = logging.getLogger(__name__)


class InvalidPackagePath(Exception):
    pass


class InvalidSermosConfig(Exception):
    pass


class MissingSermosConfig(Exception):
    pass


class InvalidCustomImage(Exception):
    pass


class ExcludeUnknownSchema(Schema):
    class Meta:
        unknown = EXCLUDE


class SermosEnvironmentVarSchema(ExcludeUnknownSchema):
    name = fields.String(
        required=True,
        description="Environment variable name.",
        example="MY_ENV_VAR"
    )
    value = fields.String(
        required=True,
        description="Environment variable value.",
        example="my special value"
    )

    @pre_load
    def redact_null_values(self, item, **kwargs):
        """ Environment variables are often secrets and therefore should not
            be committed to the sermos.yaml file. Instead, they should be
            templated with ${ENV_VAR_NAME} and injected into the file only
            in memory during deployments, etc.

            The Sermos application will use the sermos.yaml file for other
            configuration pieces that do not require the environment variables,
            therefore we add 'redacted' as the value to any null values in order
            for the schema to pass the load step as being non-null.
        """
        item["value"] = item["value"]\
            if item["value"] is not None else "redacted"
        return item


class SermosImageConfigCustomImagesSchema(ExcludeUnknownSchema):
    name = fields.String(
        required=True,
        description="Name for your image. Do not include any spaces nor "
                    "special characters.",
        example="special-feature"
    )
    command = fields.String(
        required=True,
        description="The pip install command to use.",
        example="sermos-client-pkg[core,special_feature]"
    )


class SermosImageConfigBaseImageSchema(ExcludeUnknownSchema):
    command = fields.String(
        required=True,
        description="The pip install command to use.",
        example="sermos-client-pkg[core]"
    )


class SermosImageConfigSchema(ExcludeUnknownSchema):
    baseImage = fields.Nested(
        SermosImageConfigBaseImageSchema,
        required=False
    )
    customImages = fields.List(
        fields.Nested(
            SermosImageConfigCustomImagesSchema,
            required=True
        ),
        required=False,
        description="Optional list of custom images."
    )


class SermosRouteDocumentationSchema(ExcludeUnknownSchema):
    class Meta:
        # Allows us to add some python identifiers as keys to the schema
        include = {
            'in': fields.String(
                required=True,
                description="Where the parameter exists (e.g. path, header)",
                example="path"
            ),
            'type': fields.String(
                required=True,
                description="Data type of the parameter (e.g. integer, string)",
                example="string"
            )
        }

    name = fields.String(
        required=True,
        description="Name of the parameter, typically the variable name.",
        example="apikey"
    )
    description = fields.String(
        required=False,
        description="A description of the paramter.",
        example="Your API Key"
    )
    example = fields.String(
        required=False,
        description="A useful example of the parameter",
        example="8dff67fc-7cba-49da-bc75-f29437a7d723"
    )
    required = fields.Boolean(
        required=True,
        description="Boolean flag for whether the parameter is required.",
        example="True"
    )


class SermosRouteConfigSchema(ExcludeUnknownSchema):
    route = fields.String(
        required=True,
        description="The route for this endpoint. Will be appended to base."
                    "e.g. /my-route --> /api/v1/apiUrlPrefix/my-route",
        example="/my-route"
    )
    routeDocumentation = fields.List(
        fields.Nested(
            SermosRouteDocumentationSchema,
            required=True
        ),
        required=False,
        description="Optional documentation for your route (recommended)."
    )


class SermosHttpMethodsVerbConfigSchema(ExcludeUnknownSchema):
    requestSchema = fields.String(
        required=False,
        description="The Marshmallow schema defining the request payload."
    )
    responseSchema = fields.String(
        required=False,
        description="The Marshmallow schema defining the endpoint's response."
    )


class SermosHttpMethodsConfigSchema(ExcludeUnknownSchema):
    post = fields.Nested(
        SermosHttpMethodsVerbConfigSchema,
        required=False,
        description="Request/Response schemas for the `post` method."
    )
    put = fields.Nested(
        SermosHttpMethodsVerbConfigSchema,
        required=False,
        description="Request/Response schemas for the `put` method."
    )
    get = fields.Nested(
        SermosHttpMethodsVerbConfigSchema,
        required=False,
        description="Request/Response schemas for the `get` method."
    )
    delete = fields.Nested(
        SermosHttpMethodsVerbConfigSchema,
        required=False,
        description="Request/Response schemas for the `delete` method."
    )


class SermosApiConfigEndpointsSchema(ExcludeUnknownSchema):
    handler = fields.String(
        required=True,
        description="Full path to the Class that defines this API endpoint.",
        example="sermos_customer_client.api.my_api.UsefulApiClass"
    )
    apiUrlPrefix = fields.String(
        required=True,
        description="The API prefix for this route. This is appended to the "
                    "base API route (e.g. /api/v1) and is used to group "
                    "endpoints in the API documentation. For example, "
                    "apiUrlPrefix==documents will result in a base url "
                    "of /api/v1/documents/(routeConfig.route) and will "
                    "be grouped with any other routes sharing the `documents` "
                    "prefix.",
        example="documents"
    )
    routeConfig = fields.Nested(
        SermosRouteConfigSchema,
        required=True
    )
    httpMethodsConfig = fields.Nested(
        SermosHttpMethodsConfigSchema,
        required=True
    )
    event = fields.Raw(
        required=False,
        unknown=INCLUDE,
        description="Arbitrary user data, exposed to your API endpoints "
                    "through `self.event`, which is injected by Sermos.",
    )


class SermosApiConfigPrefixDescriptionsSchema(ExcludeUnknownSchema):
    prefix = fields.String(
        required=True,
        description="The url prefix itself.",
        example="custom-prefix"
    )
    description = fields.String(
        required=True,
        description="Useful description of the url prefix.",
        example="A custom prefix to logically group operations."
    )
    name = fields.String(
        required=False,
        description="The 'nice name' of the prefix. Recommended to leave this "
                    "empty; Sermos will autogenerate from the prefix.",
        example="Custom Prefx"
    )

    @pre_load
    def create_name(self, in_data, **kwargs):
        if in_data.get('name', None) is None:
            in_data['name'] = ' '.join(in_data['prefix'].split('-')).title()
        return in_data


class SermosApiDescriptionSchema(ExcludeUnknownSchema):
    version = fields.String(
        description="The API version of your application. Recommended to leave "
                    "this field empty, Sermos will use your current package "
                    "version automatically so this is dynamically versioned.",
        example="0.1.0",
        required=False
    )
    title = fields.String(
        description="Title of your API documentation.",
        example="Company's Sermos API",
        required=False
    )
    description = fields.String(
        description="A description of your Sermos API",
        example="API endpoints managing custom document processing needs.",
        required=False
    )


class SermosApiConfigSchema(ExcludeUnknownSchema):
    customImageName = fields.String(
        required=False,
        description="Specify a custom image if not using the default. For "
                    "this to function as an API, it still needs to be an "
                    "image built by Sermos. To build more than one usable "
                    "Sermos image, you can specify multiple installations "
                    "under `imageConfig.customImages`",
        example="custom-api-name"
    )
    environmentVariables = fields.List(
        fields.Nested(
            SermosEnvironmentVarSchema,
            required=True
        ),
        description="List of name/value environment variable pairs available "
                    "to all API endpoints.",
        required=False
    )
    endpoints = fields.List(
        fields.Nested(
            SermosApiConfigEndpointsSchema,
            required=True
        ),
        description="List of endpoints to expose through your API.",
        required=True
    )
    prefixDescriptions = fields.List(
        fields.Nested(
            SermosApiConfigPrefixDescriptionsSchema,
            required=True
        ),
        description="List of custom prefixes to use in your API. This is "
                    "optional, if you exclude then you can still add new "
                    "prefixes as you like, the documentation just won't have "
                    "a description.",
        required=False
    )
    apiDocumentation = fields.Nested(
        SermosApiDescriptionSchema,
        description="Information for your API documentation service.",
        required=False
    )


class SermosRegisteredTaskDetailConfigSchema(ExcludeUnknownSchema):
    handler = fields.String(
        required=True,
        description="Full path to the Method handles work / pipeline tasks.",
        example="sermos_customer_client.workers.worker_group.useful_worker"
    )
    queue = fields.String(
        required=False,
        description="The default queue on which this worker listens. If none "
                    "is provided, it will listen to the default worker queue, "
                    "which is the default behavior (and fine in most cases). "
                    "Consider a custom queue for long running tasks and/or "
                    "specialized tasks that need extra resources/GPUs/etc. "
                    "NOTE: your Sermos deployment must specifically launch "
                    "workers that listen to any custom queues, otherwise no "
                    "tasks will be picked up.",
        example="my-long-running-task-queue"
    )
    event = fields.Raw(
        required=False,
        unknown=INCLUDE,
        description="Arbitrary user data, passed through `event` arg in task."
    )


class SermosRegisteredWorkerDetailConfigSchema(ExcludeUnknownSchema):
    name = fields.String(
        required=True,
        description="Name of the worker.",
        example="useful_worker"  # TODO add validation on this so it's valid
    )
    customImageName = fields.String(
        required=False,
        description="Specify a custom image if not using the default. For "
                    "this to function as a worker, it still needs to be an "
                    "image built by Sermos. To build more than one usable "
                    "Sermos image, you can specify multiple installations "
                    "under `imageConfig.customImages`",
        example="custom-worker-name"
    )
    queue = fields.String(
        required=False,
        description="The default queue on which this worker listens. If none "
                    "is provided, it will listen to the default worker queue, "
                    "which is the default behavior (and fine in most cases). "
                    "Consider a custom queue for long running tasks and/or "
                    "specialized tasks that need extra resources/GPUs/etc. "
                    "NOTE: your Sermos deployment must specifically launch "
                    "workers that listen to any custom queues, otherwise no "
                    "tasks will be picked up.",
        example="my-long-running-task-queue"
    )
    maxTtl = fields.Integer(
        required=False,
        description="Max TTL for a task in seconds.",
        default=300,
        example=300
    )
    replicaCount = fields.Integer(
        required=False,
        description="Number of workers to have available.",
        default=1,
        example=1
    )
    # scalingMethod = fields.String(
    #     required=False,
    #     validate=OneOf(['manual', 'queue', 'cpu'])
    #     description="Method for this worker to scale up/down.",
    #     default="manual",
    #     example="manual"
    # )
    cpuRequest = fields.String(
        required=False,
        description="CPUs expected to be available for each replica.",
        default="500m",
        example="2000m"
    )
    memoryRequest = fields.String(
        required=False,
        description="Memory expected to be available for each replica.",
        default="512Mi",
        example="2Gi"
    )
    environmentVariables = fields.List(
        fields.Nested(
            SermosEnvironmentVarSchema,
            required=True
        ),
        description="List of name/value environment variable pairs available "
                    "to this specific worker.",
        required=False
    )


class SermosWorkerConfigSchema(ExcludeUnknownSchema):
    registeredTasks = fields.List(
        fields.Nested(
            SermosRegisteredTaskDetailConfigSchema,
            required=True
        ),
        required=True,
        description="List of task handlers to register for to your Sermos app."
    )
    registeredWorkers = fields.List(
        fields.Nested(
            SermosRegisteredWorkerDetailConfigSchema,
            required=True
        ),
        required=True,
        description="List of unique workers (which can handle any number of "
                    "tasks). A worker listens to a queue. By default, there is "
                    "a generic 'work' queue that everything is placed on, "
                    "therefore, the convention is to set at least one "
                    "'base worker' that listens to the default queue. All "
                    "registeredTasks will, by default, look for their work on "
                    "the default queue. You can override this behavior by "
                    "registering the task and defining a specific queue, then, "
                    "register a worker here and specify the same queue. This "
                    "allows a quick way to separate out special types of work "
                    "that may take a long time, require more resources, have "
                    "different scaling behavior, etc."
    )
    environmentVariables = fields.List(
        fields.Nested(
            SermosEnvironmentVarSchema,
            required=True
        ),
        description="List of name/value environment variable pairs available "
                    "to all workers.",
        required=False
    )


class SermosYamlSchema(ExcludeUnknownSchema):
    name = fields.String(
        required=True,
        description="A name for your Sermos deployment.",
        example="My Deployment."
    )

    imageConfig = fields.Nested(
        SermosImageConfigSchema,
        description="An optional configuration that allows for a custom "
                    "installation command for the 'baseImage' (the "
                    "application that will run by default in all API and "
                    "Worker nodes) and, again optionally, a list of "
                    "'customImages' that can be used with any registered "
                    "worker or in your API. "
                    "Sermos will default to making a base image using a "
                    "standard `pip install my-pkg` installation. "
                    "If your python package is installed by a means other than "
                    "the standard `pip install my-pkg`, you can provide the "
                    "install command under imageConfig.baseImage.command. "
                    "Currently, this will work only with `extras`, "
                    "e.g. `my-pkg[extra1,extra2]`. "
                    "The 'customImages' key is a list of name/command pairs "
                    "used to build an image. In some cases, you need more "
                    "than one image (e.g. a special build with a particularly "
                    "large image size), this allows you to list more than one "
                    "image to build. You can use these custom images in your "
                    "workers by specifying the `customImageName` in your "
                    "`registeredWorkers` list, matching the `name` provided "
                    "in `imageConfig`.",
        required=False
    )

    environmentVariables = fields.List(
        fields.Nested(
            SermosEnvironmentVarSchema,
            required=True
        ),
        description="List of name/value environment variable pairs available "
                    "to all API classes and workers.",
        required=False
    )

    apiConfig = fields.Nested(
        SermosApiConfigSchema,
        required=False,
        description="Core API configuration."
    )

    workerConfig = fields.Nested(
        SermosWorkerConfigSchema,
        required=False,
        description="Core Worker configuration."
    )


class YamlPatternConstructor():
    """ Adds a pattern resolver + constructor to PyYaml.

        Typical/deault usage is for parsing environment variables
        in a yaml file but this can be used for any pattern you provide.
    """
    def __init__(self, env_var_pattern: str = None,
                 add_constructor: bool = True):
        self.env_var_pattern = env_var_pattern
        if self.env_var_pattern is None:
            # e.g. ${VAR:default}
            self.env_var_pattern = r'\$\{([\w\-\:\/\@]+)\}'
        self.path_matcher = re.compile(self.env_var_pattern)

        if add_constructor:
            self.add_constructur()

    def _path_constructor(self, loader, node):
        """ Extract the matched value, expand env variable, and replace the match
        """
        env_var_name = re.match(self.env_var_pattern, node.value)
        try:
            env_var_name = env_var_name.group(1)
        except AttributeError:
            return None

        env_var_name_split = env_var_name.split(':')
        env_var = os.environ.get(env_var_name_split[0], None)
        if env_var is None:
            # If a default was provided (e.g. VAR:default), return that.
            # We join anything after first element because the default
            # value might be a URL or something with a colon in it
            # which would have 'split' above
            if len(env_var_name_split) > 1:
                return ":".join(env_var_name_split[1:])
            return None  # Return None if not in environ nor default
        return env_var

    def add_constructur(self):
        """ Initialize PyYaml with ability to resolve/load environment
            variables defined in a yaml template when they exist in
            the environment.

            Add to SafeLoader in addition to standard Loader
        """
        yaml.add_implicit_resolver('!env_var', self.path_matcher)
        yaml.add_implicit_resolver(
            '!env_var', self.path_matcher, Loader=SafeLoader)
        yaml.add_constructor('!env_var', self._path_constructor)
        yaml.add_constructor(
            '!env_var', self._path_constructor, Loader=SafeLoader)


def _validate_custom_image_definitions_in_yaml(sermos_config: dict):
    """ Ensures any customImageName keys line up with customImage definitions.

        Specifically, protect against typos or missing definitions in one or
        the other. We want to prevent customImage definitions that are not
        used as well, because that would simply add to build and deploy time.
    """
    workers = sermos_config.get(
        'workerConfig', {}).get('registeredWorkers', [])
    if len(workers) == 0:
        return

    api_config = sermos_config.get('apiConfig', {})
    custom_images_list =\
        sermos_config.get('imageConfig', {}).get('customImages', [])
    custom_images = [ci['name'] for ci in custom_images_list]
    custom_images.append(DEFAULT_BASE_IMAGE_NAME)  # Add the 'base' image

    # 1) Ensure there is a corresponding customImage for every registered worker
    # custom worker
    for worker in workers:
        worker_img_name = worker.get(
            'customImageName', DEFAULT_BASE_IMAGE_NAME)
        if worker_img_name not in custom_images:
            raise InvalidCustomImage("`{}` not defined in customImages."
                                     .format(worker_img_name))

    # 2) Ensure there is a corresponding registered worker OR Api Config
    # for every customImage
    for ci in custom_images:
        registered_found = False
        for worker in workers:
            if ci == worker.get('customImageName', DEFAULT_BASE_IMAGE_NAME):
                registered_found = True
                break
        if sermos_config.get('apiConfig', {}).get('customImageName') == ci:
            registered_found = True

        if not registered_found:
            raise InvalidCustomImage("`{}` not used in any registeredTask"
                                     .format(ci))


def parse_config_file(sermos_yaml: str):
    """ Parse the `sermos.yaml` file when it's been loaded.

        Arguments:
            sermos_yaml (required): String of loaded sermos.yaml file.
    """
    YamlPatternConstructor()  # Add our env variable parser
    try:
        sermos_yaml_schema = SermosYamlSchema()
        # First suss out yaml issues
        sermos_config = yaml.safe_load(sermos_yaml)
        # Then schema issues
        sermos_config = sermos_yaml_schema.load(sermos_config)
        # Ensure, when customImages are used, that the image name exists
        _validate_custom_image_definitions_in_yaml(sermos_config)
        # TODO validate entries in imageConfig to ensure it has std chars, etc
    except ValidationError as e:
        msg = "Invalid Sermos configuration due to {}"\
            .format(e.messages)
        logger.error(msg)
        raise InvalidSermosConfig(msg)
    except InvalidCustomImage as e:
        msg = "Invalid customImage configuration due to {}"\
            .format(e)
        logger.error(msg)
        raise InvalidCustomImage(msg)
    except Exception as e:
        msg = "Invalid Sermos configuration, likely due to invalid "\
            "YAML formatting ..."
        logger.exception("{} {}".format(msg, e))
        raise InvalidSermosConfig(msg)
    return sermos_config


def load_sermos_config(pkg_name: str = None,
                       sermos_yaml_filename: str = None,
                       as_dict: bool = True):
    """ Load and parse the `sermos.yaml` file. Issue usable exceptions for
        known error modes so bootstrapping can handle appropriately.

        Arguments:
            pkg_name (required): Directory name for your Python
                package. e.g. my_package_name . If none provided, will check
                environment for `SERMOS_CLIENT_PKG_NAME`. If not found,
                will exit.
            sermos_yaml_filename (optional): Relative path to find your
                `sermos.yaml` configuration file. Defaults to `sermos.yaml`
                which should be found inside your `pkg_name`
            as_dict (optional): If true (default), return the loaded sermos
                configuration as a dictionary. If false, return the loaded
                string value of the yaml file.
    """
    sermos_config = None
    sermos_yaml = None
    if sermos_yaml_filename is None:
        sermos_yaml_filename = DEFAULT_YAML_NAME

    try:
        sermos_config_path = pkg_resources.resource_filename(
            normalized_pkg_name(pkg_name), sermos_yaml_filename
        )
    except Exception as e:
        msg = "Either pkg_name ({}) or sermos_yaml_filename ({}) is "\
            "invalid ...".format(pkg_name, sermos_yaml_filename)
        logger.error("{} ... {}".format(msg, e))
        raise InvalidPackagePath(e)

    try:
        with open(sermos_config_path, 'r') as f:
            sermos_yaml = f.read()
            sermos_config = parse_config_file(sermos_yaml)
    except InvalidSermosConfig as e:
        raise
    except InvalidCustomImage as e:
        raise
    except FileNotFoundError as e:
        msg = "Sermos config file could not be found at path {} ...".format(
            sermos_config_path
        )
        raise MissingSermosConfig(msg)
    except Exception as e:
        raise e
    if as_dict:
        return sermos_config
    return yaml.safe_dump(sermos_config)



import re
import ast
import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    user_options = [
        ('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import coverage
        import pytest

        if self.pytest_args and len(self.pytest_args) > 0:
            self.test_args.extend(self.pytest_args.strip().split(' '))
            self.test_args.append('tests/')

        cov = coverage.Coverage()
        cov.start()
        errno = pytest.main(self.test_args)
        cov.stop()
        cov.report()
        cov.html_report()
        print("Wrote coverage report to htmlcov directory")
        sys.exit(errno)


_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('sermos_utils/__init__.py', 'rb') as f:
    __version__ = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))

with open('requirements.txt', 'r') as f:
    install_requires = f.read().splitlines()


setup(
    name='sermos-utils',
    version=__version__,
    description="Sermos Utilities",
    long_description=open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    maintainer="Rho AI",
    license="MIT",
    url="https://bitbucket.org/rhoai/sermos-utils",
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    install_requires=install_requires,
    extras_require={
        'dev': [
            'honcho',
            'twine'
        ],
        'test': [
            'pytest>=5.2.4,<6',
            'tox>=3.14.1,<4',
            'coverage>=4.5,<5',
            'mock>=1,<2',
            'responses>=0.10.7,<11',
            'moto>=1.3'
        ]
    },
    cmdclass={'test': PyTest},
    entry_points="""
    [console_scripts]
    sermos_validate=sermos_utils.cli.deploy:validate
    sermos_deploy=sermos_utils.cli.deploy:deploy
    sermos_status=sermos_utils.cli.deploy:status
    """,
)



""" Utilities for deploying applications to Sermos.

    Example CLI Usage:

        honcho run -e .env sermos_deploy
        --> Pipeline UUID: {abc-123}
        honcho run -e .env sermos_status {abc-123}

    Example Programmatic Usage:

        from sermos_utils.deploy import SermosDeploy

        sd = SermosDeploy(
            os.environ.get("SERMOS_DEPLOY_KEY", None),
            pkg_name="sermos_demo_client"
        )

        # To Invoke
        status = sd.invoke_deployment()
        print(status)

        # To Check Status:
        status = sd.get_deployment_status(pipeline_uuid)
        print(status)
"""
import json
import subprocess
import base64
import logging
import requests
from sermos_utils.utils import get_deploy_key, get_client_pkg_name
from sermos_utils.sermos_yaml import load_sermos_config
from sermos_utils.constants import DEFAULT_DEPLOY_URL

logger = logging.getLogger(__name__)


class SermosDeploy():
    """ Primary Sermos Deployment class for invocation and status updates.
    """
    def __init__(self, deploy_key: str = None, pkg_name: str = None,
                 sermos_yaml_filename: str = None, commit_hash: str = None,
                 deploy_url: str = None, deploy_branch: str = 'master'):
        """ Arguments:
                deploy_key (optional): Deployment key, issued by Sermos, that
                    dictates the environment into which this request will be
                    deployed. Defaults to checking the environment for
                    `SERMOS_DEPLOY_KEY`. If not found, will exit.
                pkg_name (optional): Directory name for your Python
                    package. e.g. my_package_name . If none provided, will check
                    environment for `SERMOS_CLIENT_PKG_NAME`. If not found,
                    will exit.
                sermos_yaml_filename (optional): Relative path to find your
                    `sermos.yaml` configuration file. Defaults to `sermos.yaml`
                    which should be found inside your `pkg_name`
                commit_hash (optional): The specific commit hash of your git
                    repo to deploy. If not provided, then current HEAD as of
                    invocation will be used. This is the default usage, and is
                    useful in the case of a CI/CD pipeline such that the Sermos
                    deployment is invoked after your integration passes.
                deploy_url (optional): Defaults to primary sermos deployment
                    endpoint. Only modify this if there is a specific, known
                    reason to do so.
        """
        super(SermosDeploy, self).__init__()
        self.deploy_key = get_deploy_key(deploy_key)
        self.pkg_name = get_client_pkg_name(pkg_name)
        self.sermos_yaml_filename = sermos_yaml_filename
        self.commit_hash = commit_hash
        self.sermos_yaml = None  # Established later, only on `invoke`
        self.encoded_sermos_yaml = None  # Established later, only on `invoke`
        self.deploy_payload = None  # Established later, only on `invoke`
        self.deploy_url = deploy_url if deploy_url\
            else DEFAULT_DEPLOY_URL
        self.deploy_branch = deploy_branch
        self.headers = {
            'Content-Type': 'application/json',
            'apikey': self.deploy_key
        }

    def _set_commit_hash(self):
        """ Retrieve the commit hash of the current git state and set to
            current deployment object.
        """
        if self.commit_hash is None:
            self.commit_hash = subprocess.check_output(
                ["git", "rev-parse", "--verify", "HEAD"]
            ).strip().decode('utf-8')

    def _set_encoded_sermos_yaml(self):
        """ Provide the b64 encoded sermos.yaml file as part of request.
            Primarily used to get the custom workers definitions, etc. so
            the deployment endpoint can generate the values.yaml.
        """
        self.sermos_yaml = load_sermos_config(
            self.pkg_name, self.sermos_yaml_filename, as_dict=False)
        self.encoded_sermos_yaml = base64.b64encode(
            self.sermos_yaml.encode('utf-8')
        ).decode('utf-8')

    def _set_deploy_payload(self):
        """ Set the deployment payload correctly.
        """
        self._set_commit_hash()
        self._set_encoded_sermos_yaml()
        self.deploy_payload = {
            "sermos_yaml": self.encoded_sermos_yaml,
            "commit_hash": self.commit_hash,
            "deploy_branch": self.deploy_branch,
        }

    def get_deployment_status(self):
        """ Info on a specific deployment
        """
        this_url = self.deploy_url + '/status'
        r = requests.get(this_url, headers=self.headers)
        return r

    def validate_deployment(self, output_file: str = None):
        """ Test rendering sermos.yaml and validate.
        """
        # Running this will raise an exception if something is invalid.
        self._set_deploy_payload()

        if output_file:
            with open(output_file, 'w') as f:
                f.write(self.sermos_yaml)
        return True

    def invoke_deployment(self):
        """ Invoke a Sermos AI Deployment

            If no commit_hash was provided, use the "current" commit hash
            of the client package during this invocation.

            Required convention is that your client's python package
            version number is specified in the file `my_package/__init__.py`
            and is defined as a string assigned to the variable `__version__`,
            e.g. `__version__ = '0.1.0'`
        """
        self._set_deploy_payload()

        # Make request to your environment's endpoint
        r = requests.post(
            self.deploy_url,
            headers=self.headers,
            data=json.dumps(self.deploy_payload)
        )
        return r
